print("import: 'trafaret'")
import trafaret

print("import: 'trafaret.constructor'")
import trafaret.constructor

print("import: 'trafaret.keys'")
import trafaret.keys

print("import: 'trafaret.utils'")
import trafaret.utils

print("import: 'trafaret.contrib.rfc_3339'")
import trafaret.contrib.rfc_3339

